from contextlib import suppress
from unittest.mock import MagicMock, Mock, call, patch

from django.test import TestCase, override_settings

from django_kafka.conf import SETTINGS_KEY, settings
from django_kafka.consumer import Consumer, Topics


class StopWhileTrueException(Exception):
    pass


class ConsumerTestCase(TestCase):
    @patch(
        "django_kafka.consumer.Consumer.process_message",
        side_effect=StopWhileTrueException(),
    )
    @patch(
        "django_kafka.consumer.ConfluentConsumer",
        **{
            # first iteration - no message,
            # second iteration - 1 message represented as boolean value for testing
            "return_value.poll.side_effect": [None, True],
        },
    )
    def test_start(self, mock_consumer_client, mock_process_message):
        class SomeConsumer(Consumer):
            topics = MagicMock()
            config = {}

        consumer = SomeConsumer()

        # hack to break infinite loop
        # `consumer.start` is using `while True:` loop which never ends
        # in order to test it we need to break it which is achievable with side_effect
        with suppress(StopWhileTrueException):
            consumer.start()

        # subscribed to the topics defined by consumer class
        mock_consumer_client.return_value.subscribe.assert_called_once_with(
            topics=list(consumer.topics),
        )
        # poll for message with the frequency defined by consumer class
        self.assertEqual(
            mock_consumer_client.return_value.poll.call_args_list,
            # 2 calls expected as we setup 2 iterations in the test
            [call(timeout=consumer.polling_freq)] * 2,
        )
        # process_message called once as there was no message on the first iteration
        mock_process_message.assert_called_once_with(True)

    @patch("django_kafka.consumer.Consumer.commit_offset")
    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_process_message_success(self, mock_consumer_client, mock_commit_offset):
        class SomeConsumer(Consumer):
            topics = MagicMock()
            config = {}

        msg = Mock(error=Mock(return_value=False))

        consumer = SomeConsumer()

        consumer.process_message(msg)

        # check if msg has error before processing
        msg.error.assert_called_once_with()
        # Topic.consume called
        consumer.topics[msg.topic()].consume.assert_called_once_with(msg)
        # commit_offset triggered
        mock_commit_offset.assert_called_once_with(msg)

    @patch("django_kafka.consumer.Consumer.handle_error")
    @patch("django_kafka.consumer.Consumer.commit_offset")
    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_process_message_msg_error_logged(
        self,
        mock_consumer_client,
        mock_commit_offset,
        handle_error,
    ):
        class SomeConsumer(Consumer):
            topics = MagicMock()
            config = {}

        msg = Mock(error=Mock(return_value=True))

        consumer = SomeConsumer()

        consumer.process_message(msg)

        # check if msg has error before processing
        msg.error.assert_called_once_with()
        # error handler was triggered
        handle_error.assert_called_once_with(msg.error.return_value)
        # Topic.consume is not called
        consumer.topics[msg.topic()].consume.assert_not_called()
        # Consumer.commit_offset is not called
        mock_commit_offset.assert_not_called()

    @patch("django_kafka.consumer.Consumer.handle_error")
    @patch("django_kafka.consumer.Consumer.commit_offset")
    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_process_message_exception(
        self,
        mock_consumer_client,
        mock_commit_offset,
        handle_error,
    ):
        topic_consume_side_effect = TypeError("test")
        topic = Mock(
            **{
                "name": "topic",
                "consume.side_effect": topic_consume_side_effect,
            },
        )
        msg = Mock(**{"topic.return_value": topic.name, "error.return_value": False})

        class SomeConsumer(Consumer):
            topics = Topics(
                topic,
            )
            config = {}

        consumer = SomeConsumer()

        consumer.process_message(msg)

        # check if msg has error before processing
        msg.error.assert_called_once_with()
        # Topic.consume was triggered
        consumer.topics[msg.topic()].consume.assert_called_once_with(msg)
        # error handler was triggered on exception
        handle_error.assert_called_once_with(topic_consume_side_effect)
        # Consumer.commit_offset is not called
        mock_commit_offset.assert_not_called()

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_auto_offset_false(self, mock_consumer_client):
        class SomeConsumer(Consumer):
            config = {}

        consumer = SomeConsumer({"enable.auto.offset.store": False})

        msg = Mock()
        consumer.commit_offset(msg)

        mock_consumer_client.return_value.store_offsets.assert_called_once_with(msg)

    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_auto_offset_true(self, mock_consumer_client):
        class SomeConsumer(Consumer):
            config = {}

        consumer = SomeConsumer({"enable.auto.offset.store": True})

        consumer.commit_offset(Mock())

        mock_consumer_client.return_value.store_offsets.assert_not_called()

    def test_settings_are_correctly_assigned(self):
        self.assertEqual(Consumer.polling_freq, settings.POLLING_FREQUENCY)
        self.assertEqual(Consumer.default_error_handler, settings.ERROR_HANDLER)

    @override_settings(
        **{
            SETTINGS_KEY: {
                "CLIENT_ID": "client.id-initial-definition",
                "GLOBAL_CONFIG": {
                    "client.id": "client-id-overridden-by-global-config",
                    "bootstrap.servers": "defined-in-global-config",
                },
                "CONSUMER_CONFIG": {
                    "bootstrap.servers": "bootstrap.servers-overridden-by-consumer",
                    "group.id": "group.id-defined-by-consumer-config",
                },
            },
        },
    )
    @patch("django_kafka.consumer.ConfluentConsumer")
    def test_init_config_merge_override(self, mock_consumer_client):
        """
        1. CLIENT_ID is added to the consumers config
        2. GLOBAL_CONFIG overrides CLIENT_ID (client.id) if it contains one
        3. CONSUMER_CONFIG merges with GLOBAL_CONFIG and overrides keys if any
        4. Consumer.config is merged with CONSUMER_CONFIG and overrides keys if any
        4. Lastly `config` provided to Consumer.__init__ merges and overrides everything
        """

        class SomeConsumer(Consumer):
            config = {
                "group.id": "group.id.overridden-by-consumer-class",
                "enable.auto.offset.store": True,
            }

        # Consumer.config is properly merged
        consumer = SomeConsumer()

        self.assertDictEqual(
            consumer.config,
            {
                "client.id": "client-id-overridden-by-global-config",
                "bootstrap.servers": "bootstrap.servers-overridden-by-consumer",
                "group.id": "group.id.overridden-by-consumer-class",
                "enable.auto.offset.store": True,
            },
        )

        # config provided to the Consumer.__init__ is properly merged
        consumer = SomeConsumer({"enable.auto.offset.store": False})

        self.assertDictEqual(
            consumer.config,
            {
                "client.id": "client-id-overridden-by-global-config",
                "bootstrap.servers": "bootstrap.servers-overridden-by-consumer",
                "group.id": "group.id.overridden-by-consumer-class",
                "enable.auto.offset.store": False,
            },
        )



from unittest.mock import call, patch

from confluent_kafka.serialization import MessageField
from django.test import TestCase, override_settings

from django_kafka.conf import SETTINGS_KEY
from django_kafka.exceptions import DjangoKafkaError
from django_kafka.topic import AvroTopic, Topic


class SomeTopic(Topic):
    name = "some-topic"

    def consume(self, msg):
        pass


class TopicTestCase(TestCase):
    def setUp(self):
        self.topic = SomeTopic()

    @patch("django_kafka.topic.Topic.serialize")
    @patch("django_kafka.DjangoKafka.producer")
    def test_produce_only_value(self, mock_kafka_producer, mock_topic_serialize):
        value = "message value"
        headers = None  # default is None when not provided

        self.topic.produce(value)

        mock_topic_serialize.assert_called_once_with(value, MessageField.VALUE, headers)
        mock_kafka_producer.produce.assert_called_once_with(
            self.topic.name,
            mock_topic_serialize.return_value,
        )

    @patch("django_kafka.topic.Topic.serialize")
    @patch("django_kafka.DjangoKafka.producer")
    def test_produce_key_is_serialized(self, mock_kafka_producer, mock_topic_serialize):
        value = "message value"
        key = "some key"
        headers = {"header-1": "header-1-value"}

        self.topic.produce(value, key=key, headers=headers)

        self.assertEqual(
            mock_topic_serialize.call_args_list,
            [
                call(key, MessageField.KEY, headers),
                call(value, MessageField.VALUE, headers),
            ],
        )

        mock_kafka_producer.produce.assert_called_once_with(
            self.topic.name,
            mock_topic_serialize.return_value,
            key=mock_topic_serialize.return_value,
            headers=headers,
        )

    @patch("django_kafka.topic.Topic.key_deserializer")
    @patch("django_kafka.topic.Topic.context")
    def test_deserialize_key(self, mock_topic_context, mock_key_deserializer):
        value = b"some key"
        field = MessageField.KEY

        self.topic.deserialize(value, field)

        mock_topic_context.assert_called_once_with(field, None)
        mock_key_deserializer.assert_called_once_with(
            value,
            mock_topic_context.return_value,
        )

    @patch("django_kafka.topic.Topic.value_deserializer")
    @patch("django_kafka.topic.Topic.context")
    def test_deserialize_value(self, mock_topic_context, mock_value_deserializer):
        value = b"some value"
        field = MessageField.VALUE

        self.topic.deserialize(value, field)

        mock_topic_context.assert_called_once_with(field, None)
        mock_value_deserializer.assert_called_once_with(
            value,
            mock_topic_context.return_value,
        )

    @patch("django_kafka.topic.Topic.key_deserializer")
    @patch("django_kafka.topic.Topic.value_deserializer")
    @patch("django_kafka.topic.Topic.context")
    def test_deserialize_unknown_field(
        self,
        mock_topic_context,
        mock_value_deserializer,
        mock_key_deserializer,
    ):
        field = "something_unknown"

        with self.assertRaisesMessage(
            DjangoKafkaError,
            f"Unsupported deserialization field {field}.",
        ):
            self.topic.deserialize("some value", field)

        mock_topic_context.assert_not_called()
        mock_value_deserializer.assert_not_called()
        mock_key_deserializer.assert_not_called()

    @patch("django_kafka.topic.Topic.key_serializer")
    @patch("django_kafka.topic.Topic.context")
    def test_serialize_key(self, mock_topic_context, mock_key_serializer):
        value = "some key"
        field = MessageField.KEY

        self.topic.serialize(value, field)

        mock_topic_context.assert_called_once_with(field, None)
        mock_key_serializer.assert_called_once_with(
            value,
            mock_topic_context.return_value,
        )

    @patch("django_kafka.topic.Topic.value_serializer")
    @patch("django_kafka.topic.Topic.context")
    def test_serialize_value(self, mock_topic_context, mock_value_serializer):
        value = "some value"
        field = MessageField.VALUE

        self.topic.serialize(value, field)

        mock_topic_context.assert_called_once_with(field, None)
        mock_value_serializer.assert_called_once_with(
            value,
            mock_topic_context.return_value,
        )

    @patch("django_kafka.topic.Topic.key_serializer")
    @patch("django_kafka.topic.Topic.value_serializer")
    @patch("django_kafka.topic.Topic.context")
    def test_serialize_unknown_field(
        self,
        mock_topic_context,
        mock_value_serializer,
        mock_key_serializer,
    ):
        field = "something_unknown"

        with self.assertRaisesMessage(
            DjangoKafkaError,
            f"Unsupported serialization field {field}.",
        ):
            self.topic.serialize("some value", field)

        mock_topic_context.assert_not_called()
        mock_value_serializer.assert_not_called()
        mock_key_serializer.assert_not_called()

    def test_context(self):
        fields = [MessageField.VALUE, MessageField.KEY]
        headers = {"header-1": "header-1-value"}

        for field in fields:
            with patch(
                "django_kafka.topic.SerializationContext",
            ) as mock_serialization_context:
                self.topic.context(field, headers)
                mock_serialization_context.assert_called_once_with(
                    self.topic.name,
                    field,
                    headers=headers,
                )


class ATopic(AvroTopic):
    name = "some_topic_with_avro_serialization"

    def consume(self, msg):
        pass


@override_settings(
    **{SETTINGS_KEY: {"SCHEMA_REGISTRY": {"url": "http://schema-registy"}}},
)
@patch("django_kafka.topic.kafka.schema_client")
class AvroTopicTestCase(TestCase):
    def setUp(self):
        self.topic = ATopic()

    def test_key_schema(self, mock_kafka_schema_client):
        schema = self.topic.key_schema

        # return value of the schema.get_latest_version method call
        self.assertEqual(
            schema,
            mock_kafka_schema_client.get_latest_version.return_value,
        )
        # get_latest_version called with right arguments
        mock_kafka_schema_client.get_latest_version.assert_called_once_with(
            f"{self.topic.name}-key",
        )

    def test_value_schema(self, mock_kafka_schema_client):
        schema = self.topic.value_schema
        # return value of the schema.get_latest_version method call
        self.assertEqual(
            schema,
            mock_kafka_schema_client.get_latest_version.return_value,
        )
        # called with right arguments
        mock_kafka_schema_client.get_latest_version.assert_called_once_with(
            f"{self.topic.name}-value",
        )

    @patch("django_kafka.topic.AvroSerializer")
    def test_key_serializer(self, mock_avro_serializer, mock_kafka_schema_client):
        key_serializer = self.topic.key_serializer

        # returns AvroSerializer instance
        self.assertEqual(key_serializer, mock_avro_serializer.return_value)
        # instance was initialized with right arguments
        mock_avro_serializer.assert_called_once_with(
            mock_kafka_schema_client,
            schema_str=self.topic.key_schema.schema.schema_str,
        )

    @patch("django_kafka.topic.AvroDeserializer")
    def test_key_deserializer(self, mock_avro_deserializer, mock_kafka_schema_client):
        key_deserializer = self.topic.key_deserializer

        # returns mock_AvroDeserializer instance
        self.assertEqual(key_deserializer, mock_avro_deserializer.return_value)
        # instance was initialized with right arguments
        mock_avro_deserializer.assert_called_once_with(
            mock_kafka_schema_client,
            schema_str=self.topic.key_schema.schema.schema_str,
        )

    @patch("django_kafka.topic.AvroSerializer")
    def test_value_serializer(self, mock_avro_serializer, mock_kafka_schema_client):
        value_serializer = self.topic.value_serializer

        # returns AvroSerializer instance
        self.assertEqual(value_serializer, mock_avro_serializer.return_value)
        # instance was initialized with right arguments
        mock_avro_serializer.assert_called_once_with(
            mock_kafka_schema_client,
            schema_str=self.topic.key_schema.schema.schema_str,
        )

    @patch("django_kafka.topic.AvroDeserializer")
    def test_value_deserializer(self, mock_avro_deserializer, mock_kafka_schema_client):
        value_deserializer = self.topic.value_deserializer

        # returns mock_AvroDeserializer instance
        self.assertEqual(value_deserializer, mock_avro_deserializer.return_value)
        # instance was initialized with right arguments
        mock_avro_deserializer.assert_called_once_with(
            mock_kafka_schema_client,
            schema_str=self.topic.key_schema.schema.schema_str,
        )



import socket

from django.conf import settings as django_settings

SETTINGS_KEY = "DJANGO_KAFKA"
DEFAULTS = {
    "CLIENT_ID": f"{socket.gethostname()}-python",
    "ERROR_HANDLER": "django_kafka.error_handlers.ClientErrorHandler",
    "GLOBAL_CONFIG": {},
    "PRODUCER_CONFIG": {},
    "CONSUMER_CONFIG": {},
    "POLLING_FREQUENCY": 1,  # seconds
    "SCHEMA_REGISTRY": {},
}


class Settings:
    @property
    def _settings(self):
        return getattr(django_settings, SETTINGS_KEY, {})

    def __getattr__(self, attr):
        if attr in self._settings:
            return self._settings[attr]

        if attr in DEFAULTS:
            return DEFAULTS[attr]

        raise AttributeError(f"Invalid setting: '{attr}'")


settings = Settings()



from confluent_kafka import KafkaError

from django_kafka.exceptions import DjangoKafkaError


class ClientErrorHandler:
    def __call__(self, error: KafkaError):
        if error.fatal():
            raise DjangoKafkaError(error)



import logging
from typing import Optional

from django.core.management.base import BaseCommand

from django_kafka import kafka

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Start python consumers in parallel."

    def add_arguments(self, parser):
        parser.add_argument(
            "consumers",
            nargs="*",
            type=str,
            default=None,
            help="Python path to the consumer class(es). Starts all if not provided.",
        )

    def handle(self, consumers: Optional[list[str]] = None, *args, **options):
        kafka.start_consumers(consumers)



from functools import wraps
from typing import TYPE_CHECKING, Type

from django_kafka.exceptions import DjangoKafkaError

if TYPE_CHECKING:
    from django_kafka.consumer import Consumer


class ConsumersRegistry:
    def __init__(self):
        self.__consumers: dict[str, Type[Consumer]] = {}

    def __call__(self):
        @wraps(self)
        def add_to_registry(consumer_cls: Type["Consumer"]) -> Type["Consumer"]:
            self.__consumers[self.get_key(consumer_cls)] = consumer_cls
            return consumer_cls

        return add_to_registry

    def __getitem__(self, key: str):
        try:
            return self.__consumers[key]
        except KeyError as error:
            raise DjangoKafkaError(f"Consumer `{key}` is not registered.") from error

    def __iter__(self):
        yield from self.__consumers.keys()

    def get_key(self, consumer_cls: Type["Consumer"]) -> str:
        return f"{consumer_cls.__module__}.{consumer_cls.__name__}"



import logging
from abc import ABC, abstractmethod
from typing import Optional

from confluent_kafka import cimpl
from confluent_kafka.schema_registry.avro import AvroDeserializer, AvroSerializer
from confluent_kafka.serialization import (
    Deserializer,
    MessageField,
    SerializationContext,
    Serializer,
    StringDeserializer,
    StringSerializer,
)

from django_kafka import kafka
from django_kafka.exceptions import DjangoKafkaError

logger = logging.getLogger(__name__)


class Topic(ABC):
    key_serializer: Serializer = StringSerializer()
    key_deserializer: Deserializer = StringDeserializer()

    value_serializer: Serializer = StringSerializer()
    value_deserializer: Deserializer = StringDeserializer()

    @property
    @abstractmethod
    def name(self) -> str:
        """Define Kafka topic name"""

    @abstractmethod
    def consume(self, msg: cimpl.Message):
        """Implement message processing"""

    def produce(self, value: any, **kwargs):
        headers = kwargs.get("headers")

        if "key" in kwargs:
            kwargs["key"] = self.serialize(kwargs["key"], MessageField.KEY, headers)

        kafka.producer.produce(
            self.name,
            self.serialize(value, MessageField.VALUE, headers),
            **kwargs,
        )

    def deserialize(
        self, value, field: MessageField, headers: Optional[dict | list] = None,
    ):
        if field == MessageField.VALUE:
            return self.value_deserializer(
                value,
                self.context(MessageField.VALUE, headers),
            )

        if field == MessageField.KEY:
            return self.key_deserializer(value, self.context(MessageField.KEY, headers))

        raise DjangoKafkaError(f"Unsupported deserialization field {field}.")

    def serialize(
        self, value, field: MessageField, headers: Optional[dict | list] = None,
    ):
        if field == MessageField.VALUE:
            return self.value_serializer(
                value,
                self.context(MessageField.VALUE, headers),
            )

        if field == MessageField.KEY:
            return self.key_serializer(value, self.context(MessageField.KEY, headers))

        raise DjangoKafkaError(f"Unsupported serialization field {field}.")

    def context(
        self,
        field: MessageField,
        headers: Optional[dict | list] = None,
    ) -> SerializationContext:
        return SerializationContext(self.name, field, headers=headers)


class AvroTopic(Topic):
    @property
    def key_schema(self):
        return kafka.schema_client.get_latest_version(f"{self.name}-key")

    @property
    def value_schema(self):
        return kafka.schema_client.get_latest_version(f"{self.name}-value")

    @property
    def key_serializer(self):
        return AvroSerializer(
            kafka.schema_client,
            schema_str=self.key_schema.schema.schema_str,
        )

    @property
    def key_deserializer(self):
        return AvroDeserializer(
            kafka.schema_client,
            schema_str=self.key_schema.schema.schema_str,
        )

    @property
    def value_serializer(self):
        return AvroSerializer(
            kafka.schema_client,
            schema_str=self.value_schema.schema.schema_str,
        )

    @property
    def value_deserializer(self):
        return AvroDeserializer(
            kafka.schema_client,
            schema_str=self.value_schema.schema.schema_str,
        )
